from treepath.path.builder.dash_path_builder import DashPathRoot
from treepath.path.builder.patch_constants import recursive
from treepath.path.builder.patch_constants import recursive as rec
from treepath.path.builder.patch_constants import wildcard
from treepath.path.builder.patch_constants import wildcard as wc
from treepath.path.builder.root_path_builder import RootPathBuilder
from treepath.path.traverser.match import Match
from treepath.path.traverser.traverser_functions import find
from treepath.path.traverser.traverser_functions import find_matches
from treepath.path.traverser.traverser_functions import get
from treepath.path.traverser.traverser_functions import get_match
from treepath.path.traverser.traverser_functions import has
from treepath.path.traverser.traverser_functions import nested_find_matches
from treepath.path.traverser.traverser_functions import nested_get_match

path = RootPathBuilder()
pathd = DashPathRoot()



import os

from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.vertex.vertex import Vertex


class PathSyntaxError(TreepathException, SyntaxError):
    """PathSyntaxError is raised when there is an syntax issue with the path"""

    def __init__(self, vertex: Vertex, error_msg):
        self.error_msg = error_msg
        super().__init__(vertex)

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"""{self.error_msg}{os.linesep}  path: {path}"""



import functools
from typing import Union, Iterator, Callable, Any

from treepath.path.builder.path_builder import PathBuilder, get_vertex_from_path_builder
from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.path_predicate import PathPredicate
from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.traverser.match import Match
from treepath.path.traverser.match_traverser import MatchTraverser
from treepath.path.traverser.nested_match_traverser import NestedMatchTraverser
from treepath.path.traverser.nested_value_traverser import NestedValueTraverser
from treepath.path.traverser.value_traverser import ValueTraverser
from treepath.path.util.decorator import pretty_rep
from treepath.path.util.function import tuple_iterable

_not_set = dict()


def get(expression: PathBuilder, data: Union[dict, list, Match], default=_not_set
        ) -> Union[dict, list, str, int, float, bool, None]:
    must_match = (default is _not_set)
    match = get_match(expression, data, must_match=must_match)
    if match:
        return match.data
    return default


def find(expression: PathBuilder, data: Union[dict, list, Match]) -> Iterator[
    Union[dict, list, str, int, float, bool, None]]:
    if isinstance(data, Match):
        return nested_find(expression, data)

    vertex = get_vertex_from_path_builder(expression)
    traverser = ValueTraverser(data, vertex)
    traverser_iter = iter(traverser)
    return traverser_iter


def get_match(expression: PathBuilder, data: Union[dict, list, Match], must_match: bool = True) -> Union[Match, None]:
    if isinstance(data, Match):
        return nested_get_match(expression, data, must_match)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise MatchNotFoundError(vertex)
    return None


def find_matches(expression: PathBuilder, data: Union[dict, list, Match]) -> Iterator[Match]:
    if isinstance(data, Match):
        return nested_find_matches(expression, data)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_find(expression: PathBuilder, parent_match: Match) -> Iterator[
    Union[dict, list, str, int, float, bool, None]]:
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedValueTraverser(parent_match._traverser_match, vertex)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_get_match(expression: PathBuilder, parent_match: Match, must_match: bool = True) -> Union[Match, None]:
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise MatchNotFoundError(vertex)
    return None


def nested_find_matches(expression: PathBuilder, parent_match: Match) -> Iterator[Match]:
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex)
    traverser_iter = iter(traverser)
    return traverser_iter


def has(
        path: Union[PathBuilderPredicate, PathPredicate],
        *single_arg_functions: [Callable[[Any], Any]]) -> Callable[[Match], Any]:
    if isinstance(path, PathPredicate):
        real_path = path.path
        single_arg_operation = path.operation
    else:
        real_path = path
        single_arg_operation = None

    match_iter = functools.partial(nested_find_matches, real_path)

    if single_arg_operation and single_arg_functions:
        @pretty_rep(
            lambda: f"has({real_path} {single_arg_operation}, {', '.join(tuple_iterable(single_arg_functions))})")
        def has_predicate(parent_match: Match):
            for next_match in match_iter(parent_match):

                value = next_match.data
                for function in single_arg_functions[::-1]:
                    value = function(value)

                value = single_arg_operation(value)

                if value:
                    return True
            return False

        return has_predicate

    if not single_arg_operation and single_arg_functions:
        @pretty_rep(lambda: f"has({real_path}, {', '.join(tuple_iterable(single_arg_functions))})")
        def has_predicate(parent_match: Match):
            for next_match in match_iter(parent_match):

                value = next_match.data
                for function in single_arg_functions[::-1]:
                    value = function(value)

                if value:
                    return True
            return False

        return has_predicate

    if single_arg_operation and not single_arg_functions:
        @pretty_rep(lambda: f"has({real_path} {single_arg_operation})")
        def has_predicate(parent_match: Match):
            for next_match in match_iter(parent_match):
                value = single_arg_operation(next_match.data)
                if value:
                    return True
            return False

        return has_predicate

    else:
        @pretty_rep(lambda: f"has({real_path})")
        def has_predicate(parent_match: Match):
            for next_match in match_iter(parent_match):
                return True
            return False

        return has_predicate



from typing import Callable


def do_nothing(value):
    """
    Return the argument
    return the argument value
    """
    return value


def pretty_rep(pretty_rep_: Callable[[], str]):
    class PrettyRep:
        __slots__ = '_operator', '_pretty_rep'

        def __init__(self, operator_):
            self._operator = operator_
            self._pretty_rep = pretty_rep_

        def __call__(self, arg):
            return self._operator(arg)

        def __repr__(self):
            return self._pretty_rep()

    return PrettyRep



from typing import Union, Callable, Match, Any

from treepath.path.builder.abstract_class_builder import AbstractPathBuilder
from treepath.path.builder.patch_constants import wildcard
from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.symbol import Symbol
from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.vertex.key_vertex import KeyVertex, KeyWildVertex
from treepath.path.vertex.list_vertex import ListIndexVertex, ListSliceVertex, ListWildVertex
from treepath.path.vertex.parent_vertex import ParentVertex
from treepath.path.vertex.predicate_vertex import PredicateVertex
from treepath.path.vertex.recursive_vertex import RecursiveVertex
from treepath.path.vertex.tuple_vertex import TupleVertex
from treepath.path.vertex.vertex import Vertex

_RESERVED_ATTR_FOR_VERTEX_DATA = "_RESERVED_ATTR_FOR_VERTEX_DATA"
"""
  _RESERVED_ATTR_FOR_VERTEX_DATA:  A special attribute name used by PathBuilder to store all the DATA related to the 
  vertex.   It purpose keep data attribute separate from PathBuilder in order to reduce attribute name collision.  
"""


def _build_key(parent_vertex: Vertex, key: Union[int, slice, Symbol, str, tuple, Callable[[Match], Any]]):
    if isinstance(key, int):
        return ListIndexVertex(parent_vertex, key)
    elif isinstance(key, slice):
        return ListSliceVertex(parent_vertex, key)
    if wildcard == key:
        return ListWildVertex(parent_vertex)
    elif isinstance(key, str):
        return KeyVertex(parent_vertex, key)
    elif isinstance(key, tuple):
        return TupleVertex(parent_vertex, key)
    elif callable(key):
        return PredicateVertex(parent_vertex, key)
    else:
        raise PathSyntaxError(parent_vertex,
                              f" [{type(key)}] indices must be int, slice, tuple, str, "
                              f"wildcard, has(function), or callable")


class PathBuilder(PathBuilderPredicate, AbstractPathBuilder):
    __slots__ = _RESERVED_ATTR_FOR_VERTEX_DATA

    def __init__(self, vertex, ):
        object.__setattr__(self, _RESERVED_ATTR_FOR_VERTEX_DATA, vertex)

    def __getattr__(self, name):
        """
        creates a new attribute at self.
        This method only get called if self.name does not already exist
        self.name
        """
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = KeyVertex(parent_vertex, self.tranform_attribute_name(name))
        path_builder = self.create_path_builder(vertex)
        return path_builder

    def __setattr__(self, name, value):
        """
        Overloads self.name = value
        Attribute assignment not supported.
        """
        raise AttributeError(f"set __setattr__  {name} {value} not supported on {self.__name__}")

    def __getitem__(self, key: Union[int, slice, Symbol, str, tuple, Callable[[Match], Any]]):
        """
        Overloads self[key]
        """
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = _build_key(parent_vertex, key)
        path_builder = self.create_path_builder(vertex)
        return path_builder

    def __setitem__(self, key, value):
        """
        Overloads self[key]=value
        """
        raise AttributeError(f"set __setitem__  {key} not supported on {self.__name__}")

    def __repr__(self):
        vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        return repr(vertex)

    def __str__(self):
        return self.__repr__()

    def __dir__(self):
        pass

    def create_path_builder(self, *args, **kwargs):
        return PathBuilder(*args, **kwargs)

    def tranform_attribute_name(self, name):
        return name

    @property
    def recursive(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = RecursiveVertex(parent_vertex)
        if isinstance(parent_vertex, RecursiveVertex):
            raise PathSyntaxError(
                parent_vertex,
                "Successive recursive vertices are not allowed in the path expression."
            )
        path_builder = self.create_path_builder(vertex)
        return path_builder

    @property
    def rec(self):
        return self.recursive

    @property
    def wildcard(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = KeyWildVertex(parent_vertex)
        path_builder = self.create_path_builder(vertex)
        return path_builder

    @property
    def wc(self):
        return self.wildcard

    @property
    def parent(self):
        parent_vertex = object.__getattribute__(self, _RESERVED_ATTR_FOR_VERTEX_DATA)
        vertex = ParentVertex(parent_vertex)
        path_builder = self.create_path_builder(vertex)
        return path_builder


def get_vertex_from_path_builder(path_builder: PathBuilder) -> Vertex:
    vertex = object.__getattribute__(path_builder, _RESERVED_ATTR_FOR_VERTEX_DATA)
    return vertex



from treepath.path.exceptions.traversing_error import TraversingError


class InfiniteLoopDetected(TraversingError):
    """InfiniteLoopDetected is raised when the next is detect to be in a infinite loop."""




def create_partial_operator(operator_, right):
    def partial_operator_(left):
        return operator_(left, right)

    return partial_operator_


def tuple_iterable(tuple_: tuple):
    for entry in tuple_:
        yield str(entry)
